package hipi.compiler.generator;

import hipi.compiler.*;

import hipi.compiler.util.StringMap;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;


/**
 * Generates output files (one for each schedule).
 * 
 * @author Paolo Milazzo
 */
public class CodeGenerator implements HiPiCompilerVisitor {

	private PrintWriter outFile;
	private String encoding = "UTF8";
	private String schemaURI = "http://www.cs.unibo.it/~milazzo/hipi/hipi.xsd";

	// this is used to store informations about recursive definitions. It maps
	// names to RecDefsDescriptors
	private StringMap recDefs;
	// this contains descriptors for global channels
	private StringMap globalChannels;
	// this contains nodes of the tree related to global names (for each schedule)
	private HashSet scheduleNames;
	// this contains spawn nodes encountered during tree visit
	private StringMap spawnedThreads;

	// a counter used for thread names generation
	private static int threadCounter = 0;

	// thread names generator
	private static String generateThreadName() {
		return "___thread___" + threadCounter++;
	}
	
	public Object visit(ASTHiPiProgram node, Object data) {
		recDefs = new StringMap();
		globalChannels = new StringMap();
		StateInfo info = new StateInfo();
		node.childrenAccept(this,info);	
		return null;
	}

	public Object visit(ASTGlobalChannelDeclaration node, Object data) {
		globalChannels.put(node.name,node);
		return null;
	}

	public Object visit(ASTRecursiveDefinition node, Object data) {
		if (!((StateInfo) data).inSchedule)
			// inserts (only) global recursive defs in recDefs
			recDefs.put(node.name, 
				new RecDefsDescriptor(node,true));
		else if (!((StateInfo) data).scheduleRecDefsFirstVisit) 
		{
			((StateInfo) data).inRecDef = true;
			outFile.println("<thread name=\"" + node.name + "\">");

			outFile.println("<env idx=\"0\"/>");
			outFile.println("<env idx=\"1\"/>");
			
			// TODO this is wrong.. leads to an exception if we have overriding
			if (!((RecDefsDescriptor) recDefs.get(node.name)).isGlobal) {
				Iterator schedNamesIterator = scheduleNames.iterator();
				while (schedNamesIterator.hasNext()) {
					outFile.println("<env idx=\"" +
					((ASTLocalDeclarationStatement) schedNamesIterator.next()).envIndex +
					"\"/>");
				}
			}
			
			// copy return channel
			outFile.println("<store idx=\"2\">");
			outFile.println("<load src=\"1\"/>");
			outFile.println("</store>");
			
			// receive parameters
			outFile.println("<recv>");
			outFile.println("<load src=\"0\"/>");
			for (int i=0; i<node.params.size(); i++) {
				outFile.println("<store idx=\"" + 
					((FormalParamDescriptor) node.params.get(i)).envIndex +
					"\"/>");
			}
			outFile.println("</recv>");
			
			node.childrenAccept(this,data);
			
			// a default return statement is added
			outFile.println("<send>");
			outFile.println("<load src=\"2\"/>");
			outFile.println("</send>");
			outFile.println("<terminate/>");

			// thread end			
			outFile.println("</thread>");
			outFile.println();
			
			((StateInfo) data).inRecDef = false;
		}
		return null;
	}

	public Object visit(ASTSchedule node, Object data) {
		((StateInfo) data).inSchedule = true;
		((StateInfo) data).scheduleRecDefsFirstVisit = true;
		scheduleNames = new HashSet();
		spawnedThreads = new StringMap();
		Iterator recDefsIterator = recDefs.entrySet().iterator();
		RecDefsDescriptor descriptor;
		while (recDefsIterator.hasNext()) {
			descriptor = (RecDefsDescriptor)((Map.Entry) recDefsIterator.next()).getValue();
			descriptor.needed = false;
		}
		try {
			
			outFile = new PrintWriter(new OutputStreamWriter(new FileOutputStream(node.name + ".xml"),encoding),true);
			outFile.println("<?xml version=\"1.0\" encoding=\"" + 
					(encoding.compareTo("UTF8")== 0 ? "UTF-8" : encoding) + "\"?>");
			outFile.println("<schedule name=\"" + node.name + "\"");	
			outFile.println("          xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\""); 
			outFile.println("          xsi:noNamespaceSchemaLocation=\"" + schemaURI + "\">");
			outFile.println();

			// inserts schedule wide scope names in scheduleNames
			// and recurisive defs names in recDefs
			node.childrenAccept(new CodeGeneratorFirstVisitor(), 
								new CodeGeneratorData(scheduleNames,recDefs));
			node.childrenAccept(this,data);

			((StateInfo) data).scheduleRecDefsFirstVisit = false;			
			
			// write recDefs threads 
			recDefsIterator = recDefs.entrySet().iterator();
			Vector markedForRemove = new Vector(); // resolve ConcurrentModificationException 
			while (recDefsIterator.hasNext()) {
				descriptor = (RecDefsDescriptor)((Map.Entry) recDefsIterator.next()).getValue();
// TODO write recdefs only if needed
//				if (descriptor.needed) {
					descriptor.node.jjtAccept(this,data);
					descriptor.needed=false;
//				}
				if (!descriptor.isGlobal) markedForRemove.add(descriptor.node.name);
			}
			
			// write spawned threads
			Vector spawnedThreadsNames = new Vector(spawnedThreads.keySet());
			Vector spawnedThreadsNamesOld = new Vector();
			ASTSpawnStatement current; String threadName;
			
			((StateInfo) data).inSpawnedThread = true;
			
			// fix point iteration (for nested spawns)
			// TODO terrible performances
			while (spawnedThreadsNames.size() != spawnedThreadsNamesOld.size()) {
			
				for (int i = 0; i<spawnedThreadsNames.size() ; i++) {
					if (!spawnedThreadsNamesOld.contains(spawnedThreadsNames.get(i))) {
						current = (ASTSpawnStatement) spawnedThreads.get(spawnedThreadsNames.get(i));
				
						outFile.println("<thread name=\"" + spawnedThreadsNames.get(i) + "\">");
						outFile.println("<env idx=\"all\"/>");
				
						current.childrenAccept(this,data);
				
						outFile.println("</thread>");
						outFile.println();
					}
				}
			
				spawnedThreadsNamesOld = spawnedThreadsNames;
				spawnedThreadsNames = new Vector(spawnedThreads.keySet());
			}
			
			((StateInfo) data).inSpawnedThread = false;

			// clean recDefs
			while (!markedForRemove.isEmpty())
				recDefs.remove(markedForRemove.remove(0));
						
			outFile.println("</schedule>");
									
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		((StateInfo) data).inSchedule = false;
		
		return null;
	}
	
	public Object visit(ASTScheduleMain node, Object data) {
		((StateInfo) data).inMain = true;
		outFile.println("<thread name=\"main\">");
		
		// code generation for schedule names initialization
		Iterator schedNamesIterator = scheduleNames.iterator();
		SimpleNode current;
		while (schedNamesIterator.hasNext()) {
			current = (SimpleNode) schedNamesIterator.next();
		if (ASTLocalDeclarationStatement.class.isInstance(current)) {
			if ((((ASTLocalDeclarationStatement)current).type == Types.CHANNEL) && 
				(ASTNewChannel.class.isInstance(current.jjtGetChild(0)))) 
					// optimization
					outFile.println("<newch idx=\"" + 
						((ASTLocalDeclarationStatement)current).envIndex + "\"/>");
				else {
					outFile.println("<store idx=\"" + 
						((ASTLocalDeclarationStatement) current).envIndex + "\">");
					current.childrenAccept(this,data);
					outFile.println("</store>");
				}
			}
			else if (ASTRecursiveDefinitionCall.class.isInstance(current)) {
				current.childrenAccept(this,data);
			}
		}
		node.childrenAccept(this,data);
		
		outFile.println("</thread>");
		outFile.println();
		
		((StateInfo) data).inMain = false;
		return null;
	}

	public Object visit(ASTBlock node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTLocalDeclarationStatement node, Object data) {
		if (!node.CG_visited) {
			if (ASTNewChannel.class.isInstance(node.jjtGetChild(0))) 
				// new channel assignment: optimization
				outFile.println("<newch idx=\"" + node.envIndex + "\"/>");
			else {
				outFile.println("<store idx=\"" + node.envIndex + "\">");
				node.childrenAccept(this,data);
				outFile.println("</store>");
			}
		}
		return null;
	}

	public Object visit(ASTSend node, Object data) {
		if (globalChannels.containsKey(node.name)) {
			outFile.println("<send>");
			outFile.println("<load src=\"" + 
				((ASTGlobalChannelDeclaration)globalChannels.get(node.name)).uri +
				"\"/>");
		}
		else {
			outFile.println("<send>");
			outFile.println("<load src=\"" + node.envIndex + "\"/>");
		}
		node.childrenAccept(this,data);
		outFile.println("</send>");
		return null;
	}

	public Object visit(ASTRecv node, Object data) {

		if (globalChannels.containsKey(node.name)) {
			outFile.println("<recv>");
			outFile.println("<load src=\"" + 
				((ASTGlobalChannelDeclaration)globalChannels.get(node.name)).uri +
				"\"/>");
		}
		else { 
			outFile.println("<recv>");
			outFile.println("<load src=\"" + node.envIndex + "\"/>");
		}
		int envIndex;
		for (int i=0; i<node.params.size(); i++) {
			envIndex = ((FormalParamDescriptor) node.params.get(i)).envIndex;
			// for recdefs encoding 
			if (envIndex != -1)
				outFile.println("<store idx=\"" + envIndex + "\"/>");
			else
				outFile.println("<discard/>");
		}
		outFile.println("</recv>");
		return null;
	}

	public Object visit(ASTRecursiveDefinitionCall node, Object data) {
		if (!node.CG_visited) {
			node.childrenAccept(this,data);
		}
		return null;
	}

	public Object visit(ASTSpawnStatement node, Object data) {
		if ((!node.CG_visited) && (!node.nameGiven)){
			node.threadName = generateThreadName();
			spawnedThreads.put(node.threadName,node);
			node.CG_visited = true;
			node.nameGiven = true;
		}		

		outFile.print("<spawn thread=\"" + node.threadName + "\"");
		if (node.isMigration) {
			if (globalChannels.containsKey(node.dest)) {
				outFile.print(" dest=\"" + 
					((ASTGlobalChannelDeclaration)globalChannels.get(node.dest)).uri + 
					"\"");
			}
			else {
				outFile.print(" dest=\"" + node.envIndex + "\"");				
			}
		}
		outFile.println("/>");
		if (recDefs.containsKey(node.threadName)) {
			((RecDefsDescriptor)recDefs.get(node.threadName)).needed = true;
		}
		return null;
	}

	public Object visit(ASTSelectionStatement node, Object data) {
		outFile.println("<if>");
		// for debug 
		// TODO: better debug messages printer
		if ((node.jjtGetNumChildren() != 2) &&
			(node.jjtGetNumChildren() != 3)) 
			System.err.println("DEBUG: WRONG IF STM GENERATION");
			
		node.jjtGetChild(0).jjtAccept(this,data); // boolean expression
		outFile.println("<then>");
		node.jjtGetChild(1).jjtAccept(this,data); // first block (then)
		outFile.println("</then>");
		if (node.jjtGetNumChildren() == 2) {
			outFile.println("<else/>");
		}
		else {
			outFile.println("<else>");
			node.jjtGetChild(2).jjtAccept(this,data); // second block (else)
			outFile.println("</else>");
		}
		outFile.println("</if>");
		return null;
	}

	public Object visit(ASTExpOp node, Object data) {
		outFile.println("<op type=\"" + node.operator + "\">");
		node.childrenAccept(this,data); 
		outFile.println("</op>");
		return null;
	}

	public Object visit(ASTNewChannel node, Object data) {
		outFile.println("<newch/>");
		node.childrenAccept(this,data); // no children expected
		return null;
	}

	public Object visit(ASTReturnStatement node, Object data) {
		if (((StateInfo) data).inMain || ((StateInfo) data).inSpawnedThread)
			outFile.println("<terminate/>");
		else {
			outFile.println("<send>");
			outFile.println("<load src=\"2\"/>");
			node.childrenAccept(this, data);
			outFile.println("</send>");
			outFile.println("<terminate/>");
		}
		return null;
	}
	
	public Object visit(ASTExpIdentifier node, Object data) {
		if (node.envIndex != -1)
			outFile.println("<load src=\"" + node.envIndex + "\"/>");
		else
			outFile.println("<load src=\"" + 
				((ASTGlobalChannelDeclaration) globalChannels.get(node.name)).uri +
				"\"/>");
		node.childrenAccept(this,data); // no children expected
		return null;
	}
	
	public Object visit(ASTExpInteger node, Object data) {
		outFile.println("<int>" + node.value + "</int>");
		node.childrenAccept(this,data); // no children expected
		return null;
	}
	
	public Object visit(ASTExpString node, Object data) {
		outFile.println("<string>" + node.value + "</string>");
		node.childrenAccept(this,data); // no children expected
		return null;
	}

	public Object visit(SimpleNode node, Object data) {
		// if node is a Forwarder
		if (ASTForwarder.class.isInstance(node)) 
			visit((ASTForwarder) node, data);
		else {
			// TODO gestire errore		
			System.out.println("ASTSimpleNode visited");
			node.dump(">");
		}
		return null;
	}

	// not present in HiPiCompilerVisitor (node not generated by jjtree)
	public Object visit(ASTForwarder node, Object data) {
		outFile.println("<fwd>");
		node.childrenAccept(this,data); // two children expected
		outFile.println("</fwd>");
		return null;
	}

	public Object visit(ASTTypeDeclaration node, Object data) {
		// TODO Auto-generated method stub
		return null;
	}

	public Object visit(ASTImportDeclaration node, Object data) {
		// TODO Auto-generated method stub
		return null;
	}

}
